from nmigen import *
from nmigen.hdl.rec import *
from nmigen.cli import main

import skidbuffer

from functools import reduce


# We know this is not really likely to synthesize well, but we implement here so
# we can get a gold-standard/reference implementation, against which we can test
# our further optimizations, including compression and non-standard representations.

# TODO PLAN:
# 1) get this working in simulation
# 2) make it synthesizable (so it can use FPGA block RAM with reasonable bit-width)
# 3) make it optimized




class TableBusLayout(Layout):
    def __init__(self, *, row_input_width, column_input_width, output_width):
        super().__init__([
            # INPUTS
            ("row_idx",     unsigned(row_input_width)),    # FROM SOURCE
            ("col_idx",     unsigned(column_input_width)), # FROM SOURCE
            ("valid_in",    1),                            # FROM SOURCE
            ("ready_in",    1),        # FROM DEST

            # OUTPUTS
            ("output_data", unsigned(output_width)), # TO DEST
            ("valid_out",   1), # TO DEST
            ("ready_out",   1), # TO SOURCE

        ])

class TableBus(Record):
    def __init__(self, *, row_input_width, column_input_width, output_width):
        super().__init__(TableBusLayout(row_input_width=row_input_width, column_input_width=column_input_width, output_width=output_width))



class LRTable(Elaboratable):
    def __init__(self, number_of_states, number_of_terminals, number_of_reduce_rules, input_array):
        # Parameters

        # So the LR table is a two-dimensional array, where one of the dimensions (rows)
        # is indexed by state number, and the other dimension (columns) is indexed by
        # terminal.

        # Each element of the array can be one of four values:
        # 1) Shift terminal
        # 2) Accept parse
        # 3) Error
        # 4) Reduce with rule number k

        # This means that the length of each element has to be:
        # TIW = (2 bits + number of bits to index the reduce rules)

        # We choose the width of the memory to be the width of each element

        # and the depth to be:
        # depth = number_of_states * number_of_terminals

        # We calculate these now:

        reduce_rule_signal = Signal(range(number_of_reduce_rules))
        reduce_rule_bits = len(reduce_rule_signal)
        self.table_width = (reduce_rule_bits + 2)
        self.table_depth = number_of_states * number_of_terminals

        self.number_of_terminals = number_of_terminals



        dummy_row    = Signal(range(number_of_states))
        dummy_column = Signal(range(number_of_terminals))

        # Interfaces
        self.bus = TableBus(row_input_width=len(dummy_row),
                            column_input_width=len(dummy_column),
                            output_width=self.table_width)
        # Interfaces



        # Prepare the table for consumption
        assert(len(input_array) == number_of_states)

        rasterized = []
        for row in input_array:
            assert(len(row) == number_of_terminals)
            rasterized.extend(row)

        # Memory
        self.mem = Memory(width=self.table_width, depth=self.table_depth, init=rasterized)



    def elaborate(self, platform):
        m = Module()
        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()


        m.d.sync += self.bus.valid_out.eq(self.bus.valid_in)
        m.d.comb += self.bus.ready_out.eq(1)

        # Now we calculate the address:
        tgt_address = Signal(self.table_depth)
        with m.If(self.bus.valid_in == 1):
            m.d.comb += tgt_address.eq(self.bus.row_idx * self.number_of_terminals + self.bus.col_idx)
            m.d.comb += rport.addr.eq(tgt_address)
        with m.If(self.bus.valid_out == 1):
            m.d.comb += self.bus.output_data.eq((rport.data))

        return m



class GOTOtable(Elaboratable):
    def __init__(self, number_of_states, number_of_nonterminals, input_array):
        # Parameters

        # So the GOTO table is a two-dimensional array, where one of the dimensions (rows)
        # is indexed by state number, and the other dimension (columns) is indexed by
        # nonterminal.

        # Each element of the array is of the form:

        # [VALID BIT] [STATE NUMBER]

        # where the valid bit is 1 if there is a valid mapping, and it is zero if there is
        # no valid mapping. Naturally we can optimize this in later implementations, but here
        # we accept an extra bit to keep everything simple

        # This means that the length of each element has to be:
        # TIW = (1 bit + number of bits to index the nonterminals)

        # Because this is not initially meant for synthesis efficiency, we
        # initially just choose the width of the memory to be:

        # width = number_of_terminals * TIW

        # and the depth to be:
        # depth = number_of_states

        # We calculate these now:

        nonterminal_signal = Signal(range(number_of_nonterminals))
        nonterminal_bits = len(nonterminal_signal)
        self.table_width = (nonterminal_bits + 1)
        print(number_of_nonterminals)
        self.table_depth = number_of_nonterminals * number_of_states

        self.number_of_nonterminals = number_of_nonterminals



        dummy_row    = Signal(range(number_of_states))
        dummy_column = Signal(range(number_of_nonterminals))

        # Interfaces
        self.bus = TableBus(row_input_width=len(dummy_row),
                            column_input_width=len(dummy_column),
                            output_width=self.table_width)


        # Prepare the table for consumption
        assert(len(input_array) == number_of_states)

        rasterized = []
        for row in input_array:
            assert(len(row) == number_of_nonterminals)
            rasterized.extend(row)

        # Memory

        self.mem = Memory(width=self.table_width, depth=self.table_depth, init=rasterized)


    def elaborate(self, platform):
        m = Module()
        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()


        m.d.sync += self.bus.valid_out.eq(self.bus.valid_in)
        m.d.comb += self.bus.ready_out.eq(1)

        # Now we calculate the address:
        tgt_address = Signal(self.table_depth)
        with m.If(self.bus.valid_in == 1):
            m.d.comb += tgt_address.eq(self.bus.row_idx * self.number_of_nonterminals + self.bus.col_idx)
            m.d.comb += rport.addr.eq(tgt_address)
        with m.If(self.bus.valid_out == 1):
            m.d.comb += self.bus.output_data.eq(rport.data)



        return m




class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.table = table = LRTable(3,3,3,[[1,2,3],[4,5,6],[1,9,1]])
        counter = Signal(8)
        m.d.sync += counter.eq(counter+1)

        with m.If(counter == 3):
            m.d.comb += table.bus.valid_in.eq(1)
            m.d.comb += table.bus.row_idx.eq(0)
            m.d.comb += table.bus.col_idx.eq(0)
        with m.If(counter == 4):
            m.d.comb += table.bus.valid_in.eq(1)
            m.d.comb += table.bus.row_idx.eq(1)
            m.d.comb += table.bus.col_idx.eq(2)
        with m.If(counter == 6):
            m.d.comb += table.bus.valid_in.eq(1)
            m.d.comb += table.bus.row_idx.eq(2)
            m.d.comb += table.bus.col_idx.eq(1)
        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
