import collections
import functools
import random
from enum import *
import networkx as nx


# Rather than re-implement every *LR(1) table generation algorithm (and have it output non/semi-serialized
# states), our approach is to use an external *LR(1) table generation algorithm, ingest its generated tables,
# which will be fully serialized -- the state machine only looks at the top element of the stack, and the
# state number therein is a serialized representation of the language items below -- and then generates
# fully non-serialized tables.

# The non-serialized tables do not have any sort of serialization of state numbers -- each state in the
# deterministic pushdown automata is identified by the top N items on the stack. The current version
# stores the stack in flipflops, allowing parallel access to all the items on the stack, which is fed into
# a fully-combinatorial rule matcher that determines what actions (shift, reduce, error, accept) the parser
# should take as a function of the presented input.

# Eschewing state serialization allows the FPGA IP to avoid storing large tables in block RAM, which is
# scarce. Instead, the grammar's rules are effectively encoded in the logic. For instance, if the grammar
# has a rule that looks like:

# TERM -> OPENPAREN EXPRESSION CLOSEPAREN

# This FPGA parser will have logic that looks for an OPENPAREN, followed by an EXPRESSION, and then a
# CLOSEPAREN on the stack. Examining the stack to find potential candidates for reduction rule application
# is known in the literature as "handle finding". With our implementation, no serialization is needed,
# and no indexing of states is needed -- just a simple parallel examination of the stack to find out what
# reduce rule we can apply (or if we cannot apply any reduce rule, what characters can be shifted).

# While the current implementation decisions (stack fully stored in FPGA flipflops, fully combinatorial rule
# matcher) are not optimal, they serve as a proof of concept that this sort of parser can be
# parametrized with an arbitrary grammar, integrated with other gateware, and rigorously tested in
# simulation,

# Unlike the situation with the conventional implementation strategy of LR parser (fully serialized states,
# only examining the top item on the stack, large tables), there is significant room for improvement.
# The areas of improvement include:
# 1) Store the parse stack partially in RAM, in order to reduce number of flipflops used
# 2) pipelining/simplification of complex combinatorial logic.

# The handle-finding task is effectively a regular-expression/string-matching task, and we note that there
# are a family of algorithms that efficiently implement those tasks on FPGAs / parallel hardware.

# Interestingly, using a more sophisticated handle-finding algorithm may also allow moving part of
# the parse stack into RAM, decreasing usage of flipflops. Indeed, statefulness in the handle-finder
# means the handle-finder can "remember" what it saw, even if those language items disappear into RAM.
# We note that this seems like a resurgence of the serialized indices in traditional LR parser
# implementation, and this is indeed the case. However, decoupling the handle-finding automaton
# from the LR parse tables allows us to implement it however we wish, using the full capabilities
# of our highly-parallel hardware.

# Indeed, there's no reason we cannot implement the handle-finder partly in fully-combinatorial
# logic that operates on language items in flipflops and partly with pipelined/stateful
# matching of items as they are written into a stack in RAM. 

# For instance, some level of serialization is useful since stack elements far down the stack
# may affect the current parser behavior, but since they do not change often, it is wasteful to
# store them in flip-flops (compared to less-expensive on-die or off-die RAM). It is also wasteful
# routing wires that constantly re-read those rarely-changing items and send them to logic that
# re-evaluates, every clock cycle, the same operation on the same data.

# However, combinatorial handle-finding has its advantages as well. Pipelined/sophisticated string
# matching algorithms are generally designed for ingesting a continuous sequence of data, and this
# does not always hold in this use case. Shifts are fine, shifts just add data to the sequence that
# the string matching algorithm ingests. However, every reduce operation rewrites part of the stack!
# If the pipelined string matcher has ingested a portion of the stack that has been rewritten, its
# state will be invalid and the corresponding registers must be flushed *and* reinitialized (if they
# cannot be rewritten in-place) for the pipelined string matcher to produce correct results.

# Anyway, onto the algorithm.



# I don't know if this is going to actually work. Please forgive me.


# We represent fully-baked LR(1) parametrization as so:

# The shift/reduce table is an array (indexed by state number)
#                              of dictionaries (indexed by lookahead (always a terminal symbol)


# The GOTO table is an array (indexed by state number)
#                      of dictionaries (indexed by produced nonterminal)


# The entries in those tables are tuples, the first entry of the tuple is an enum which indicates
# shift/reduce/goto, the second is the state number


SRAction = Enum("Shift/Reduce Table Action", "SHIFT REDUCE ACCEPT")

GOTOAction = Enum("Goto Table Action", "GOTO")



# We use the expression grammar as a test:

nonterminals = Enum("Nonterminals", "EXPRESSION TERM FACTOR")

terminals = Enum("Terminals", "INTEGER ADDOP MULTOP OPENPAREN CLOSEPAREN ENDOFPARSE")


rules = [
         (nonterminals.EXPRESSION, [nonterminals.EXPRESSION, terminals.ADDOP,      nonterminals.TERM      ]),
         (nonterminals.EXPRESSION, [nonterminals.TERM                                              ]),
         (nonterminals.TERM,       [nonterminals.TERM,       terminals.MULTOP,     nonterminals.FACTOR    ]),
         (nonterminals.TERM,       [nonterminals.FACTOR                                            ]),
         (nonterminals.FACTOR,     [terminals.OPENPAREN,  nonterminals.EXPRESSION, terminals.CLOSEPAREN]),
         (nonterminals.FACTOR,     [terminals.INTEGER                                           ])
         ]


shift_reduce_table =\
[
{terminals.INTEGER:    (SRAction.SHIFT,  5), terminals.OPENPAREN:  (SRAction.SHIFT,  4)},    # STATE 0

{terminals.ADDOP:      (SRAction.SHIFT,  6), terminals.ENDOFPARSE: (SRAction.ACCEPT, None)}, # STATE 1

{terminals.ADDOP:      (SRAction.REDUCE, 2), terminals.MULTOP:     (SRAction.SHIFT,  7),     # STATE 2
 terminals.CLOSEPAREN: (SRAction.REDUCE, 2), terminals.ENDOFPARSE: (SRAction.REDUCE, 2)},

{terminals.ADDOP:      (SRAction.REDUCE, 4), terminals.MULTOP:     (SRAction.REDUCE, 4),     # STATE 3
 terminals.CLOSEPAREN: (SRAction.REDUCE, 4), terminals.ENDOFPARSE: (SRAction.REDUCE, 4)},

{terminals.INTEGER:    (SRAction.SHIFT,  5), terminals.OPENPAREN:  (SRAction.SHIFT,  4)},    # STATE 4

{terminals.ADDOP:      (SRAction.REDUCE, 6), terminals.MULTOP:     (SRAction.REDUCE, 6),
 terminals.CLOSEPAREN: (SRAction.REDUCE, 6), terminals.ENDOFPARSE: (SRAction.REDUCE, 6)},    # STATE 5

{terminals.INTEGER:    (SRAction.SHIFT,  5), terminals.OPENPAREN:  (SRAction.SHIFT,  4)},    # STATE 6

{terminals.INTEGER:    (SRAction.SHIFT,  5), terminals.OPENPAREN:  (SRAction.SHIFT,  4)},    # STATE 7

{terminals.ADDOP:      (SRAction.SHIFT,  6), terminals.CLOSEPAREN: (SRAction.SHIFT,  11)},   # STATE 8

{terminals.ADDOP:      (SRAction.REDUCE, 1), terminals.MULTOP:     (SRAction.SHIFT,  7),     # STATE 9
 terminals.CLOSEPAREN: (SRAction.REDUCE, 1), terminals.ENDOFPARSE: (SRAction.REDUCE, 1)},

{terminals.ADDOP:      (SRAction.REDUCE, 3), terminals.MULTOP:     (SRAction.REDUCE, 3),     # STATE 10
 terminals.CLOSEPAREN: (SRAction.REDUCE, 3), terminals.ENDOFPARSE: (SRAction.REDUCE, 3)},

{terminals.ADDOP:      (SRAction.REDUCE, 5), terminals.MULTOP:     (SRAction.REDUCE, 5),     # STATE 11
 terminals.CLOSEPAREN: (SRAction.REDUCE, 5), terminals.ENDOFPARSE: (SRAction.REDUCE, 5)},


]


goto_table =\
[
{nonterminals.EXPRESSION: (GOTOAction.GOTO, 1), nonterminals.TERM:   (GOTOAction.GOTO, 2), # STATE 0
 nonterminals.FACTOR:     (GOTOAction.GOTO, 3)},

{}, # STATE 1

{}, # STATE 2

{}, # STATE 3

{nonterminals.EXPRESSION: (GOTOAction.GOTO, 8), nonterminals.TERM:   (GOTOAction.GOTO, 2), # STATE 4
 nonterminals.FACTOR:     (GOTOAction.GOTO, 3)},

{}, # STATE 5

{nonterminals.TERM:       (GOTOAction.GOTO, 9), nonterminals.FACTOR: (GOTOAction.GOTO, 3)}, # STATE 6

{nonterminals.FACTOR:     (GOTOAction.GOTO, 10)},                                           # STATE 7

{}, # STATE 8
{}, # STATE 9
{}, # STATE 10
{}, # STATE 11
]



#print(shift_reduce_table[0][terminals.INTEGER])

for terminal in [e for e in terminals]:
    print(terminal)
    for statenumber, rule in enumerate(shift_reduce_table):
        if terminal in rule:
            print(statenumber, rule[terminal])
        else:
            print(statenumber)
    print("\n")


print("GOTO TABLE")
for nonterminal in [e for e in nonterminals]:
    print(nonterminal)
    for statenumber, rule in enumerate(goto_table):
        if nonterminal in rule:
            print(statenumber, rule[nonterminal])
        else:
            print(statenumber)
    print("\n")#for statenumber,idx in enumerate(shift_reduce_table):
#    print(statenumber, ": ", idx, "\n")



# Now we convert the tables into the graph for the deterministic pushdown automaton.


StackOp = Enum("Stack Operation", "PUSH POP")


SSD = Enum("Stack State", "DONTCARE BOTTOMOFSTACK")


sr_graph = nx.MultiDiGraph()

# First we add a node for each of the enumerated states

for statenumber, rule in enumerate(shift_reduce_table):
    sr_graph.add_node(statenumber)

# Then we add some stack-push transitions based on the shifts *alone*
for node in sr_graph.nodes():
    sr_graph.nodes[node]["label"] = (node, [])

sr_graph.nodes[0]["label"] = (0, [SSD.BOTTOMOFSTACK])


for initial_statenumber, rule in enumerate(shift_reduce_table):
    for valid_transition in rule:
        # Here, valid_transition is an index of the dictionary (so the index will be the character that causes the transition
        # and so we will label the edge with the index, and the destination and shift/reduce type is found by accessing the
        # dictionary itself)
        destination_statenumber = rule[valid_transition][1]
        if (destination_statenumber != None):
            if (rule[valid_transition][0] == SRAction.SHIFT):
                sr_graph.add_edge(initial_statenumber, destination_statenumber, label=(StackOp.PUSH, valid_transition))


nx.drawing.nx_pydot.write_dot(sr_graph, "foo.dot")


# Then we add some stack-push transitions THAT DO NOT EXIST IN THE ACTUAL LR AUTOMATON
# THIS IS PURELY SO WE CAN DETERMINE THE INVARIANT (OF STACK STATE) THAT HOLDS FOR EACH ENUMERATED STATE

for initial_statenumber, rule in enumerate(goto_table):
    for valid_transition in rule:
        # Here, valid_transition is an index of the dictionary (so the index will be the character that causes the transition
        # and so we will label the edge with the index, and the destination and shift/reduce type is found by accessing the
        # dictionary itself)
        destination_statenumber = rule[valid_transition][1]
        if (destination_statenumber != None):
            if (rule[valid_transition][0] == GOTOAction.GOTO):
                print("INITIAL STATE IS", initial_statenumber, "DESTINATION IS", destination_statenumber)
                sr_graph.add_edge(initial_statenumber, destination_statenumber, label=(StackOp.PUSH, valid_transition))



#sr_graph.add_edge(1,42,label="X")


# we implement a sort of forward-backward algorithm to determine the stack states for each of the LR-parser's stack numbers



# The stack state for each node is represented as a list.
# An empty stack is only matched by a single-element list with the pseudosymbol "BOTTOMOFSTACK":
# [SSD.BOTTOMOFSTACK]

# A stack with only a single item is matched by:
# [nonterminals.TERM, SSD.BOTTOMOFSTACK]

# After we push an EXPRESSION on that stack, it looks like

# [nonterminals.EXPRESSION, nonterminals.TERM, SSD.BOTTOMOFSTACK]
#  ^ top of stack


# A stack with a single item on the top is matched by:
# [nonterminals.TERM, SSD.DONTCARE]

# After we push an EXPRESSION on that stack, it looks like

# [nonterminals.EXPRESSION, nonterminals.TERM, SSD.DONTCARE]

# This allows to distinguish between a set of items on the top of a stack where there may be
# items below, and a set of items *that are the only items on the stack*.


# When a node has an extant stack state description and we wish to update it with information, we
# follow these rules:

# current_desc = [X_0, X_1, ..., X_n]
# new_desc     = [Y_0, Y_1, ..., Y_n]

# We examine pairs (X_i, Y_i), starting at i=0, the top of the stack. If both of them
# are real symbols and are identical, the updated stack state description
# has that symbol in that index.

# If both of them are real symbols (nonterminal or terminal) and differ, the updated
# stack state description has a DONTCARE there and that's it, there's no more entries
# in the stack state description.

# If one of them is a real symbol and the other one is a pseudosymbol (BOTTOMOFSTACK or DONTCARE),
# this reduces to a DONTCARE.

# If both of them are a DONTCARE, then, well, DONTCARE.
# If there's an BOTTOMOFSTACK and a DONTCARE, then it's a DONTCARE.
# If they're both an BOTTOMOFSTACK then it's an BOTTOMOFSTACK.

# Initially, we assign a special dummy value to each node's stack state description, representing
# that we don't know anything about it, so nothing can be derived from it.

# Only one node is initially tagged with something else than the dummy value -- the zero node,
# which is tagged with [BOTTOMOFSTACK]

# We then repeatedly iterate over edges, letting information on stack states flow through the
# edges (generated by SHIFT entries in the SHIFT/REDUCE table *and* by GOTO entries in the
# GOTO table). We stop when a new round yields no changes for every node.



# the above case analysis is implemented here
def unify(old_stack_state, new_stack_state):
    print("UNIFY CALLED!", "old_stack_state is", old_stack_state, "new state is", new_stack_state)


    if (old_stack_state == []):
        return new_stack_state
    updated_state = []

    for old_elem, new_elem in zip(old_stack_state, new_stack_state):
        if (isinstance(old_elem, SSD) or isinstance(new_elem, SSD)):
            if (old_elem == SSD.BOTTOMOFSTACK and new_elem == SSD.BOTTOMOFSTACK):
                updated_state.append(SSD.BOTTOMOFSTACK)
                break
            updated_state.append(SSD.DONTCARE)
            break
        if (old_elem != new_elem):
            updated_state.append(SSD.DONTCARE)
            break
        if (old_elem == new_elem):
            updated_state.append(old_elem)
    print("UNIFY CALLED!", "updated state is ", updated_state)

    return updated_state

print("unify ", unify([3,5],[3,4]))


#for index,x in zip(range(100),sr_graph.edges()):
#    edge_op = sr_graph[x[0]][x[1]][0]["label"]
#    stack_op_from = sr_graph.nodes[x[0]]
#    stack_op_to   = sr_graph.nodes[x[1]]
#    if (stack_op_from["label"][1] != [] and stack_op_to["label"][1] == []):
#        print("egg", stack_op_from["label"][1], "XX", edge_op[1])
#        stack_op_to["label"] = (stack_op_to["label"][0], [edge_op[1]] +stack_op_from["label"][1] )


changed = 1
iterations = 0
while (changed == 1):
    changed = 0
    iterations +=1
    for index,x in zip(range(100),sr_graph.edges()):
        edge_op = sr_graph[x[0]][x[1]][0]["label"]
        stack_op_from = sr_graph.nodes[x[0]]
        stack_op_to   = sr_graph.nodes[x[1]]

        stack_state_from_shiftop = [edge_op[1]] + stack_op_from["label"][1]

        current_stack_state = stack_op_to["label"][1]

        new_stack_state = unify(current_stack_state, stack_state_from_shiftop)

        if (new_stack_state != current_stack_state):
            changed = 1

        stack_op_to["label"] = (stack_op_to["label"][0], new_stack_state)

print("it took us", iterations, "iterations to hit the fixed point")
#print(list(nx.dfs_edges(sr_graph)))

for x in sr_graph.nodes():
    print(sr_graph.nodes[x])

nx.drawing.nx_pydot.write_dot(sr_graph, "processed_stack_state_graph.dot")



# Check for anomalous weird stuff with multiple transitions from one node to another node
for x in sr_graph.edges():
    assert(len(sr_graph[x[0]][x[1]]) == 1)

# We then cross-check the generated stack state descriptions against the original shift/goto
# entries -- given the tail of each edge, the edge's head must be consistent with the label on
# the edge.

for x in sr_graph.edges():
    edge_op = sr_graph[x[0]][x[1]][0]["label"]
    stack_op_from = sr_graph.nodes[x[0]]
    stack_op_to   = sr_graph.nodes[x[1]]

    assert(edge_op[1] == stack_op_to["label"][1][0])

# We also check that the generated stack state descriptions are consistent with the reduce
# rules. Every node is associated with a set of reduce rules that can be taken from it, and
# the right-hand-side of all the reduce rules for that node must be explicitly demarcated in
# that node's stack state description.

# Here we extract the reduce entries from the shift-reduce table:
reduce_rule_locations = []

for statenumber, rule in enumerate(shift_reduce_table):
    for valid_transition in rule:
        # Here, valid_transition is an index of the dictionary (so the index will be the character that causes the transition
        # and so we will label the edge with the index, and the destination and shift/reduce type is found by accessing the
        # dictionary itself)
        if (rule[valid_transition][0] == SRAction.REDUCE):
            reduce_rule_number = rule[valid_transition][1]
            reduce_rule_locations.append((statenumber, reduce_rule_number))

# Now we check that if reduce rule n can be invoked from stack state k, that the right-hand-side of reduce rule n
# is explicitly present in the stack state description k.

for state_number, reduce_rule_number in reduce_rule_locations:
    fromgraph = sr_graph.nodes[state_number]["label"]
    assert(fromgraph[0] == state_number)

    stack_state = fromgraph[1]
    print(stack_state)
    rule_RHS = rules[reduce_rule_number - 1][1][::-1]

    for idx, x in enumerate(rule_RHS):
        assert(stack_state[idx] == x)


# Finally, we check for uniqueness of stack state descriptions. No two may be identical. It
# is ok for two to overlap, [A, B, C] and [A, B] are allowed to co-exist, they are dealt with
# the "pairwise priority rule" mechanism in the parser gateware.

def find_dupes(in_list):
    working_list = in_list.copy()
    seen_items = []
    duplicate_items = []
    while working_list != []:
        pulled_item = working_list.pop()
        if pulled_item in working_list:
            return pulled_item
    return None



all_stack_states = []



for n in sr_graph.nodes:
    all_stack_states.append(sr_graph.nodes[n]["label"][1])
#all_stack_states.append(all_stack_states[1])


dup = find_dupes(all_stack_states)
if dup != None:
    print("DUPLICATE FOUND! ERROR!", dup)
    exit (1)



# So this takes care of the "stack states".
# We still need to convert the (shift/reduce table, goto table, list of reduce rules)
# into:
#
# 1) Valid item ruleset (list of all terminal symbols acceptable for each state)
# 2) Shift ruleset (list of all terminal symbols for each state that we are to shift)
# 3) reduce ruleset (list of all terminal symbols for each state that are to provoke a reduce
#    and the corresponding rule number to invoke)
# 4) The list of reduce rules (well, specifically, how many items to pop and what is the new
#    symbol to push)
# 5) An encoding from symbolic terminals/nonterminals/pseudosymbols (like "ENDOFSTACK", "OPENPAREN", "EXPRESSION")
#    into hexadecimal constants
# Fortunately, this is the easy part :)


valid_item_ruleset = []
# array (indexed by state number) of arrays of valid items (where the valid items for that state number live)

force_shift_ruleset = []
# array (indexed by state number) of arrays of shift items(where the items that provoke a shift for that state number live)

reduce_ruleset = []
# array (indexed by state number) of arrays of tuples (each tuple is of the form (item, reduce rule number))

reduce_rule_execute_ruleset = []
# array (indexed by reduce rule number) of tuples (each tuple is of the form (number of items to pop, item to push))


# here we populate the valid item, force shift, and reduce rulesets
for statenumber, rule in enumerate(shift_reduce_table):
    this_state_s_valid_items = []
    this_state_s_force_shift_items = []
    this_state_s_reduce_rule_items = []

    for item_accepted in rule:
        this_state_s_valid_items.append(item_accepted)
        if (rule[item_accepted][0] == SRAction.SHIFT):
            this_state_s_force_shift_items.append(item_accepted)
        if (rule[item_accepted][0] == SRAction.REDUCE):
            this_state_s_reduce_rule_items.append((item_accepted,rule[item_accepted][1]))

    valid_item_ruleset.append(this_state_s_valid_items)
    force_shift_ruleset.append(this_state_s_force_shift_items)
    reduce_ruleset.append(this_state_s_reduce_rule_items)

print("XXXXXXX", len(valid_item_ruleset))


for rule in rules:
    reduce_rule_execute_ruleset.append((rule[0], len(rule[1])))

print(reduce_rule_execute_ruleset)